from typing import Callable

import attr

from src.utility import get_attacker_utility, get_torch_grad_utility, \
    get_nn_loss_function


@attr.s
class NeuralNetworkConfig:
    epochs: int = attr.ib(default=2000)
    """
    Number of epochs in a neural network training phase
    """

    deep_nn: True = attr.ib(default=True)
    """
    Set either deep nn (with 35 neurons) or shallow one (with 5 neurons)
    """

    learning_rate = 0.1e-3
    """
    Learning rate for Adam optimiser
    """

    batch_size = 1000
    """
    Size of a benign data batch used in each epoch of training
    """

    loss_function: Callable = attr.ib(init=False)
    """
    Loss function used for training
    """

@attr.s
class DefenderConfig:
    defender_epsilon: float = attr.ib(default=1e-3)
    """
    New actor's action is considered already played in the game if difference
    between utility of new action and nash equilibrium value of zero-sum game
    is less than this value
    """

    number_of_nn_to_train: int = attr.ib(default=1)
    """
    This number of neural networks will be trained in each double oracle
    iteration and the best one will be considered as a best response
    """

    nn_conf: NeuralNetworkConfig = attr.ib(default=NeuralNetworkConfig())
    """
    conf of neural networks
    """

    benign_data_file_name: str = attr.ib(default='normal_distribution_2D_experiments.csv')
    """
    Name of .csv file located in src/data/scored directory with feature points
    which will be used as benign data in neural network training phase.
    Space of these features must be the same as the feature space defined in
    root config.
    """


@attr.s
class AttackerConfig:

    use_gradient_descent: bool = attr.ib(default=True)
    """
    If set to True gradient descent within pytorch.optim package
    is used to find attacker's best response
    If set to False, attacker actions are discrete and the whole space is
    traversed to find best response. Example for R^2:
    [(.0,.01),(.0,.02),...,(.1,.1)]
    """

    epsion: float = attr.ib(default=1e-3)
    """
    New actor's action is considered already played in the game if difference
    between utility of new action and nash equilibrium value of zero-sum game
    is less than this value
    """

    tries_for_best_response: int = attr.ib(default=264)
    """
    Number of random tries to find attacker action using gradient descent.
    The one with best final loss value would be chosen.
    Attention. Used only when use_gradient_descent is set to True!
    """

    learning_rate = 0.5e-4
    """
    Learning rate for optimiser which updates attacker action while searching
    for best response using gradient descent
    Attention. Used only when use_gradient_descent is set to True!
    """

    epochs = 2000
    """
    Number of iterations used to update actions using gradient descent while
    searching for best response
    Attention. Used only when use_gradient_descent is set to True!
    """


@attr.s
class ModelConfig:
    features_count: int = attr.ib(default=2)
    """
    Space of feature points. Must be the same as space of feature points in data
    file specified in neural networks config
    """

    attacker_conf: AttackerConfig = attr.ib(default=AttackerConfig())
    """
    Attacker conf
    """

    defender_conf: DefenderConfig = attr.ib(default=DefenderConfig())
    """
    Defender conf
    """

    alpha_a: int = attr.ib(default=1)
    """
    alpha_a constant
    
    use method set_alpha_a to update this constant
    """

    alpha_d: int = attr.ib(default=4)
    """
    alpha_d constant
    
    use method set_alpha_d to update this constant
    """

    unified_constant: int = attr.ib(default=1)
    """
    constant unifying other constants -> (P_b * C_b) / (P_m, C_m)
    
    use method set_unified_constant to update this constant
    """

    attacker_utility: Callable = attr.ib(init=False)
    """
    Function to calculate utility for the attacker given the actions
    f: List[float], NeuralNetwork -> float
    """

    attacker_torch_utility: Callable = attr.ib(init=False)
    """
    Attacker utility function using torch tensors with gradient property
    Used for attacker to find best response via gradient descent
    """

    def __attrs_post_init__(self):
        self.attacker_utility = get_attacker_utility(self.alpha_a)
        self.attacker_torch_utility = get_torch_grad_utility(self.alpha_a)
        self.defender_conf.nn_conf.loss_function = \
            get_nn_loss_function(self.alpha_a, self.alpha_d, self.unified_constant)

    def set_alpha_a(self, alpha_a):
        self.alpha_a = alpha_a
        self.__attrs_post_init__()

    def set_alpha_d(self, alpha_d):
        self.alpha_d = alpha_d
        self.__attrs_post_init__()

    def set_unified_constant(self, unified_constant):
        self.unified_constant = unified_constant
        self.__attrs_post_init__()

    def set_data_file(self, name: str):
        self.defender_conf.benign_data_file_name = name


@attr.s
class PlotterConfig:
    enabled: bool = attr.ib(default=True)
    """
    If enabled, summarization each iteration is plotted
    """

@attr.s
class RootConfig:
    debug: bool = attr.ib(default=True)
    """
    Sets logger to debug level
    """

    model_conf: ModelConfig = attr.ib(default=ModelConfig())
    """
    Configuration of used model
    """

    plot_conf: PlotterConfig = attr.ib(default=PlotterConfig())
    """
    Information about Plotting results
    Unused for features != 2
    """