import itertools
import time
from collections import Counter
from dataclasses import dataclass
from typing import List, Tuple, Callable

import numpy as np
import pulp

from src.data.loader import np_matrix_from_scored_csv

# Global "config" variables
features_num = 2
axis_size = 50
latencies = list(np.linspace(0, 1, 5))

alpha_a = 1
alpha_d = 3
unified_constant = 5
data_file = 'normal_distribution_2D_experiments.csv'


@dataclass
class Result:
    zero_sum_game_value: float
    fp_part: float
    almost_sum_game_attacker: float
    almost_sum_game_defender: float
    time_elapsed: float

    attacker_actions: List[Tuple[float]]
    attacker_probs: List[float]

    defender_actions: List[Tuple[float]]
    defender_probs: List[List[float]]
    latencies_available: List[float]


def create_actions(just_one_axis=False):
    one_axis = np.linspace(0, 1, axis_size)
    if just_one_axis:
        return one_axis

    generator = itertools.product(*itertools.repeat(one_axis, features_num))
    return list(generator)


def utility(attacker_action: Tuple, l: float):
    return np.prod(attacker_action) * ((1-l)**alpha_a)


def solve_with_lp(actions_attacker: List,
                  actions_defender: List,
                  loss: Callable,
                  benign_data_prob: dict):
    print('Going to solve with LP')

    # Create LP problem
    m = pulp.LpProblem("Zero sum game", pulp.LpMinimize)

    # Minimizing value "v"
    v = pulp.LpVariable("v")
    m += v

    # Player two probabilities vector
    print("Defining defenders probabilities...")
    probs_defender = []
    for action in actions_defender:
        probs_point = [pulp.LpVariable(f'p({",".join(list(str(x) for x in action))},{k})', 0, 1) for k in latencies]
        m += pulp.lpSum(probs_point) == 1  # Probabilities sum to 1
        probs_defender.append(probs_point)

    fp_cost = 0
    for action, prob in zip(actions_defender, probs_defender):
        for j in range(len(latencies)):
            fp_cost += loss(latencies[j])*benign_data_prob[tuple(action)]*prob[j]
    fp_cost = fp_cost * unified_constant

    print("Defining main constraint...")
    constraints = []
    for i in range(len(actions_attacker)):
        suma = [fp_cost]
        for j in range(len(latencies)):
            suma.append(probs_defender[i][j] * utility(actions_attacker[i], latencies[j]))
        constraints.append(pulp.lpSum(suma) <= v)
    for c in constraints:
        m += c

    print("Ok, let's solve now...")
    m.solve()

    print(f'LP solved')
    print(f'Value of the game: {v.varValue}')
    print(f'Found solution: {pulp.LpStatus[m.status]}')
    print(f'Attacker\'s probabilities:')
    print(f'{list(str(abs(c.pi)) + " " for c in constraints)}')
    print(f'Deffender\'s probabilities:')
    def_probs_final = []
    for probs in probs_defender:
        print(f'{list(str(item.varValue) + " " for item in probs)}')
        def_probs_final.append([c.varValue for c in probs])

    return v.varValue, [abs(c.pi) for c in constraints], def_probs_final


def calc_fals_positives(p2_actions, p2_probs, loss, benign_probs):
    fp_cost = 0
    for action, prob in zip(p2_actions, p2_probs):
        for j in range(len(latencies)):
            fp_cost += loss(latencies[j])*benign_probs[tuple(action)]*prob[j]
    return fp_cost * unified_constant


def main():
    attacker_actions = create_actions()
    defender_actions = create_actions()

    # load data, round data and make occurrences dict
    print('preload the data first')
    benign = np_matrix_from_scored_csv(data_file, 0)[0]
    benign_data_prob = Counter()
    one_axis = create_actions(just_one_axis=True)
    for b in benign:
        idxs = []
        for num in b:
            idx = np.abs(one_axis - num).argmin()
            idxs.append(idx)

        val = tuple(one_axis[i] for i in idxs)
        benign_data_prob[val] = benign_data_prob.get(val, 0) + 1
    for key, val in benign_data_prob.items():
        benign_data_prob[key] = val/len(benign)

    # Define testing loss functions
    fp_cost_loss = lambda x: x ** alpha_d

    # Prepare timer and run solver
    start = time.time()
    res = solve_with_lp(attacker_actions, defender_actions, fp_cost_loss,
                        benign_data_prob)
    time_elapsed = time.time() - start

    game_val, attacker_probs, probs_defender = res
    fp_part = calc_fals_positives(defender_actions, probs_defender,
                                  fp_cost_loss, benign_data_prob)
    final_experiment = Result(game_val,
                              fp_part,
                              game_val - fp_part,
                              -game_val,
                              time_elapsed,
                              attacker_actions, attacker_probs,
                              defender_actions, probs_defender,
                              latencies)
    print('Experiment done.')
    return final_experiment


if __name__ == "__main__":
    result = main()
