import logging
import os

import attr
import numpy as np
import torch
from torch import nn

from src.config import NeuralNetworkConfig

logger = logging.getLogger(__name__)
DEVICE = torch.device(os.environ.get('device', 'cuda'))

@attr.s
class FormattedData:
    unique_x: np.array = attr.ib()
    probs_x: np.array = attr.ib()
    y: np.array = attr.ib()


@attr.s
class BenignData:
    unique_x: np.array = attr.ib()
    counts: np.array = attr.ib()
    y: np.array = attr.ib()


def get_nn_model(input_features: int, deep_nn: bool):
    if deep_nn:
        return nn.Sequential(
            nn.Linear(input_features, 20),
            nn.ReLU(),
            nn.Linear(20, 15),
            nn.ReLU(),
            nn.Linear(15, 20),
            nn.ReLU(),
            nn.Linear(20, 1),
            nn.Tanh(),
            SoftClip(50)
        ).to(DEVICE)
    else:
        return nn.Sequential(
            nn.Linear(input_features, 5),
            nn.ReLU(),
            nn.Linear(5, 5),
            nn.ReLU(),
            nn.Linear(5, 1),
            nn.Tanh(),
            SoftClip(50)
        ).to(DEVICE)


class OrderCounter:
    order = 0

    @staticmethod
    def next():
        OrderCounter.order += 1
        return OrderCounter.order


class SoftClip(nn.Module):
    """ SoftClipping activation function

    https://arxiv.org/pdf/1810.11509.pdf
    """
    def __init__(self, p=50.0):
        super().__init__()
        self.p = p

    def forward(self, x):
        first_pow = torch.pow(np.e, torch.mul(x, self.p))
        second_pow = torch.pow(np.e, torch.mul(torch.add(x, -1.0), self.p))

        first_div = torch.add(first_pow, 1.0)
        second_div = torch.add(second_pow, 1.0)

        division = torch.div(first_div, second_div)

        second_part_log = torch.log(division)
        first_part = 1.0 / self.p

        return torch.mul(second_part_log, first_part)


class NeuralNetwork:
    def __init__(self, input_features=2,
                 nn_conf: NeuralNetworkConfig = NeuralNetworkConfig()):
        self.model = get_nn_model(input_features, nn_conf.deep_nn)
        self._set_weights()
        self.conf = nn_conf
        self.id = OrderCounter.next()

        # Variables used for loss function
        self.attacker_actions: FormattedData = None
        self.benign_data: BenignData = None

        # Variables from last training epoch measuring quality
        self.final_loss = None
        self.final_fp_cost = None

    def __str__(self):
        return f'Neural network id:{self.id}, final loss: {self.final_loss}'

    def set_data(self, benign_data: BenignData, attack: FormattedData):
        self.attacker_actions = attack
        self.benign_data = benign_data

    def _prepare_data(self):
        defender = self.benign_data
        attacker = self.attacker_actions
        x = np.concatenate((defender.unique_x, attacker.unique_x), axis=0)
        y = np.concatenate((defender.y, attacker.y), axis=0)
        probs = np.concatenate((defender.counts/np.sum(defender.counts),
                                attacker.probs_x), axis=0)

        self.train_y = torch.cat((torch.zeros(self.conf.batch_size).float(),
                                  torch.tensor(attacker.y).float())).to(DEVICE)

        self.all_x = torch.tensor(x).float().to(DEVICE)
        self.all_y = torch.tensor(y).float().to(DEVICE)
        self.all_probs = torch.tensor(probs).float().to(DEVICE)

    def _set_weights(self):
        def init_weights(m):
            if type(m) == nn.Linear:
                torch.nn.init.xavier_uniform_(m.weight)
                m.bias.data.fill_(.0)
        self.model.apply(init_weights)

    def train(self):
        self._prepare_data()
        self._train()

    def get_train_batch(self):
        batch_idxs = np.random.choice(len(self.benign_data.unique_x),
                                      self.conf.batch_size)
        current_batch_samples = np.sum(self.benign_data.counts[batch_idxs])

        batch_x_np = np.concatenate((self.benign_data.unique_x[batch_idxs],
                                     self.attacker_actions.unique_x), axis=0)
        batch_probs_np = np.concatenate((self.benign_data.counts[batch_idxs]/current_batch_samples,
                                 self.attacker_actions.probs_x), axis=0)

        batch_x = torch.tensor(batch_x_np).float().to(DEVICE)
        batch_probs = torch.tensor(batch_probs_np).float().to(DEVICE)
        return batch_x, batch_probs

    def _train(self):
        learning_rate = self.conf.learning_rate
        optimizer = torch.optim.Adam(self.model.parameters(), lr=learning_rate)

        for e in range(self.conf.epochs):
            batch_x, batch_probs = self.get_train_batch()

            # Forward pass: compute predicted y by passing x to the model
            train_ltncies = self.latency_predict(batch_x, with_grad=True)

            # Compute loss
            loss, _ = self.conf.loss_function(batch_x, train_ltncies,
                                              self.train_y, batch_probs)

            # Log loss function value each 5 epochs
            if e % 50 == 0:
                logging.debug(f'Epoch: {e}/{self.conf.epochs},\t'
                              f'TrainLoss: {loss},\t')

            # Before the backward pass, use the optimizer object to zero all of
            # the gradients for the variables it will update
            optimizer.zero_grad()

            # Backward pass: compute gradient of the loss with respect to model
            # parameters
            loss.backward()

            # Calling the step function on an Optimizer makes an update to its
            # parameters
            optimizer.step()

        with torch.no_grad():
            train_ltncies = self.latency_predict(self.all_x, with_grad=True)
            loss, fp_part = self.conf.loss_function(self.all_x, train_ltncies,
                                                 self.all_y, self.all_probs)
            logger.debug(f'Final loss of this nn: {loss}\tfp_part is: {fp_part}')
            # measuring quality of final network
            self.final_loss = loss.item()
            self.final_fp_cost = fp_part.item()

    def _raw_predict(self, tensor: torch.Tensor):
        pred = self.model(tensor)
        return pred.flatten().float()

    def latency_predict(self, x: torch.Tensor, with_grad=False):
        if with_grad:
                raw_prediction = self._raw_predict(x)
        else:
            with torch.no_grad():
                raw_prediction = self._raw_predict(x)

        return raw_prediction

    def predict_single_latency(self, input, return_tensor=False):
        in_type = type(input)
        if in_type == list or in_type == tuple or \
                in_type == np.array or in_type == np.ndarray:
            input = torch.tensor(input).float().to(DEVICE)

        if return_tensor:
            return self.latency_predict(input)[0]
        else:
            return self.latency_predict(input)[0].item()
