import logging
from typing import List
from typing import TYPE_CHECKING

import numpy as np
import torch

# Hack to avoid cycle imports while using type checking
# The TYPE_CHECKING constant is always False at runtime
if TYPE_CHECKING:
    from src.neural_networks.network import NeuralNetwork

logger = logging.getLogger(__name__)


def get_attacker_utility(alpha_a: int):
    def attacker_utility(attacker_features: List[float], nn: 'NeuralNetwork'):
        pred = nn.predict_single_latency(attacker_features)
        return np.product(attacker_features) * (1 - pred) ** alpha_a

    return attacker_utility


def get_torch_grad_utility(alpha_a: int):
    def attacker_utility(attacker_features: torch.tensor, nn: 'NeuralNetwork'):
        latency = nn.latency_predict(attacker_features, with_grad=True)
        return torch.pow(torch.add(1, -latency), alpha_a) \
               * torch.prod(attacker_features, dim=1)

    return attacker_utility


def get_nn_loss_function(alpha_a: int, alpha_d: int, unified_constant: int):
    def loss_function(x, latencies, real_y, probs):
        zero_sum_part = torch.sum(real_y*((1-latencies)**alpha_a)*torch.prod(x, dim=1)*probs)
        fp_cost = torch.mul(torch.sum((1-real_y) * probs * torch.pow(latencies, alpha_d)), unified_constant)

        return torch.add(zero_sum_part, fp_cost), fp_cost

    return loss_function
